<?php

namespace Drupal\commerce_wechat_pay\Plugin\Commerce\PaymentGateway;

use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\commerce_price\Price;
use EasyWeChat\Payment\Order;
use Symfony\Component\HttpKernel\Exception\BadRequestHttpException;

/**
 * Provides WeChat Micropay gateway.
 * @link https://pay.weixin.qq.com/wiki/doc/api/micropay.php?chapter=5_1
 *
 * @CommercePaymentGateway(
 *   id = "micropay",
 *   label = "WeChat Micropay",
 *   display_label = "WeChat Micropay",
 *    forms = {
 *     "offsite-payment" = "Drupal\commerce_wechat_pay\PluginForm\MicropayForm",
 *   }
 * )
 */
class Micropay extends QRCodePaymentMode2 {
  use StringTranslationTrait;

  /**
   * @param string $order_id
   * @param string $auth_code
   * @param Price $total_amount
   * @return \Drupal\Core\Entity\EntityInterface
   */
  public function capture($order_id, $auth_code, Price $total_amount) {
    if (!$this->gateway_lib) {
      $this->loadGatewayConfig();
    }
    /** @var \EasyWeChat\Payment\API $gateway; */
    $gateway = $this->gateway_lib;

    $attributes = [
      'auth_code' => $auth_code,
      'body'             => \Drupal::config('system.site')->get('name') . $this->t(' Order: ') . $order_id,
      // 'detail'           => 'order_items in certain format',
      'out_trade_no'     => $order_id,
      'total_fee' => $total_amount->getNumber() * 100, // WeChat Pay use Integer for its price
      'fee_type' => $total_amount->getCurrencyCode(),
    ];

    $wechat_order = new Order($attributes);

    try {
      /** @var \EasyWeChat\Support\Collection $result */
      $response = $gateway->pay($wechat_order);

      if ($response->return_code == 'SUCCESS') {
        // Payment is successful
        $result = $response->toArray();
        if ($response->result_code == 'SUCCESS') {
          $payment_entity = $this->createPayment($result, 'capture_completed');
        } else {
          // TODO: need to handle Wait user to input password
          $payment_entity = $this->createPayment($result, 'authorization', $order_id, 'USERPAYING', $total_amount);
        }
        return $payment_entity;

      } else {
        // Payment is not successful
        \Drupal::logger('commerce_wechat_pay')->error(print_r($response->toArray(), TRUE));
        throw new BadRequestHttpException($response->return_code . '  ' .$response->return_msg);
      }
    } catch (\Exception $e) {
      // Payment is not successful
      \Drupal::logger('commerce_wechat_pay')->error($e->getMessage());
      throw new BadRequestHttpException($e->getMessage());
    }
  }
}
